from __future__ import annotations

from enum import Enum, auto
from functools import partial
from numbers import Number
from typing import TYPE_CHECKING, Any, Callable, Iterable, cast

import numpy as np
import numpy.typing as npt
from numpy import _NoValue as NoValue  # type: ignore[attr-defined]

import ch5mpy
from ch5mpy._typing import NP_FUNC
from ch5mpy.array.chunks.iter import iter_chunks_2
from ch5mpy.indexing import FullSlice, SingleIndex, map_slice

if TYPE_CHECKING:
    from ch5mpy import H5Array


class ApplyOperation(Enum):
    set = auto()
    iadd = auto()
    imul = auto()
    iand = auto()
    ior = auto()


def ensure_h5array_first(
    x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
) -> tuple[
    H5Array[Any],
    npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    Callable[[npt.NDArray[Any] | H5Array[Any], npt.NDArray[Any] | H5Array[Any]], npt.NDArray[Any] | H5Array[Any]],
    Callable[[npt.NDArray[Any] | H5Array[Any], npt.NDArray[Any] | H5Array[Any]], npt.NDArray[Any] | H5Array[Any]],
]:
    """One of x1, x2 must be an H5Array. Returns x1, x2 (maybe swapped such that the first returned array is guaranteed
    to be an H5Array).
    Also return getter functions swapping arrays back in the order they were passed to this function if needed."""
    if not isinstance(x1, ch5mpy.H5Array):
        return cast(ch5mpy.H5Array[Any], x2), x1, lambda _, x2: x2, lambda x1, _: x1

    return x1, x2, lambda x1, _: x1, lambda _, x2: x2


class MaskWhere:
    def __init__(
        self,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue,
        shape: tuple[int, ...],
    ):
        # TODO : make memory efficient (avoid broadcast and compute on the fly)
        self._where = None if where in (True, NoValue) else np.broadcast_to(where, shape)  # type: ignore[arg-type]

    def __repr__(self) -> str:
        return f"MaskWhere({self._where})"

    def __getitem__(self, item: tuple[Any, ...] | slice) -> npt.NDArray[np.bool_] | None:
        if self._where is None:
            return None

        return self._where[item]


def _get_output_array(
    out: H5Array[Any] | npt.NDArray[Any] | None,
    shape: tuple[int, ...],
    axis: tuple[int, ...],
    keepdims: bool,
    dtype: npt.DTypeLike | None,
    initial: int | float | complex | NoValue,
) -> H5Array[Any] | npt.NDArray[Any]:
    if keepdims:
        expected_shape = tuple(s if i not in axis else 1 for i, s in enumerate(shape))

    else:
        expected_shape = tuple(s for i, s in enumerate(shape) if i not in axis)

    if out is not None:
        ndim = len(expected_shape)
        if out.ndim != ndim:
            raise ValueError(f"Output array has the wrong number of dimensions: Found {out.ndim} but expected {ndim}")

        if out.shape != expected_shape:
            raise ValueError(f"Output array has the wrong shape: Found {out.shape} but expected {expected_shape}")

        if initial is not NoValue:
            out[()] = initial

        return out

    if initial is NoValue:
        return np.empty(shape=expected_shape, dtype=dtype)

    return np.full(shape=expected_shape, fill_value=initial, dtype=dtype)


def _get_output_array_2(
    out: H5Array[Any] | npt.NDArray[Any] | None,
    a_shape: tuple[int, ...],
    b_shape: tuple[int, ...],
    dtype: npt.DTypeLike | None,
    default: Any,
) -> H5Array[Any] | npt.NDArray[Any]:
    expected_shape = np.broadcast_shapes(a_shape, b_shape)

    if out is not None:
        ndim = len(expected_shape)
        if out.ndim != ndim:
            raise ValueError(f"Output array has the wrong number of dimensions: Found {out.ndim} but expected {ndim}")

        if out.shape != expected_shape:
            raise ValueError(f"Output array has the wrong shape: Found {out.shape} but expected {expected_shape}")

        return out

    if default is None:
        return np.empty(shape=expected_shape, dtype=dtype)

    return np.full(shape=expected_shape, fill_value=default, dtype=dtype)


def _as_tuple(
    axis: int | Iterable[int] | tuple[int, ...] | None,
    ndim: int,
    default_0D_output: bool,
) -> tuple[int, ...]:
    if axis is None:
        return tuple(range(ndim)) if default_0D_output else ()

    elif not isinstance(axis, Iterable):
        return (axis,)

    return tuple(axis)


def _get_indices(
    index: tuple[SingleIndex | FullSlice, ...],
    axis: tuple[int, ...],
    where_compute: MaskWhere,
    where_output: MaskWhere,
    output_ndim: int,
) -> tuple[npt.NDArray[np.bool_] | None, tuple[slice, ...], npt.NDArray[np.bool_] | None]:
    # compute on whole array at once
    if len(index) == 1 and isinstance(index[0], FullSlice) and index[0].is_whole_axis:
        return where_compute[:], (), where_output[index]

    where_to_compute = where_compute[map_slice(index)]

    # 0D output array (no chunk selection, no where selection)
    if output_ndim == 0:
        return where_to_compute, (), None

    # nD output array
    selected_index = map_slice(tuple(e for i, e in enumerate(index) if i not in axis))
    return where_to_compute, selected_index, where_output[selected_index]


def _apply_operation(
    operation: ApplyOperation,
    dest: H5Array[Any] | npt.NDArray[Any],
    chunk_selection: tuple[slice, ...] | tuple[()],
    where_to_output: npt.NDArray[np.bool_] | None,
    values: npt.NDArray[Any],
) -> None:
    if values.ndim == 0:
        values = values[()]

    if operation is ApplyOperation.set:
        if dest.ndim == 0:
            dest[()] = values

        else:
            dest[chunk_selection][where_to_output] = values[where_to_output]

    elif operation is ApplyOperation.iadd:
        if dest.ndim == 0:
            dest[()] = values + dest[()]

        else:
            dest[chunk_selection][where_to_output] += values[where_to_output]

    elif operation is ApplyOperation.imul:
        if dest.ndim == 0:
            dest *= values  # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] *= values[where_to_output]

    elif operation is ApplyOperation.iand:
        if dest.ndim == 0:
            dest &= values  # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] &= values[where_to_output]

    elif operation is ApplyOperation.ior:
        if dest.ndim == 0:
            dest |= values  # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] |= values[where_to_output]

    else:
        raise NotImplementedError(f"Do not know how to apply operation '{operation}'")


def apply(
    func: partial[NP_FUNC],
    operation: ApplyOperation,
    a: H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None,
    *,
    dtype: npt.DTypeLike | None,
    initial: int | float | complex | NoValue,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue,
    default_0D_output: bool = True,
) -> Any:
    dtype = a.dtype if dtype is None else dtype
    axis = _as_tuple(func.keywords.get("axis", None), a.ndim, default_0D_output)
    output_array = _get_output_array(out, a.shape, axis, func.keywords.get("keepdims", False), dtype, initial)

    if where is not False:
        where_compute = MaskWhere(where, a.shape)
        where_output = MaskWhere(where, output_array.shape)

        for index, chunk in a.iter_chunks(keepdims=True):
            where_to_compute, chunk_selection, where_to_output = _get_indices(
                index, axis, where_compute, where_output, output_array.ndim
            )
            result = np.array(
                func(chunk, where=True if where_to_compute is None else where_to_compute),
                dtype=output_array.dtype,
            )
            _apply_operation(operation, output_array, chunk_selection, where_to_output, result)

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array


def apply_everywhere(
    func: partial[NP_FUNC],
    operation: ApplyOperation,
    a: H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None,
    *,
    dtype: npt.DTypeLike | None,
    initial: int | float | complex | NoValue,
    default_0D_output: bool = True,
) -> Any:
    dtype = a.dtype if dtype is None else dtype
    axis = _as_tuple(func.keywords.get("axis", None), a.ndim, default_0D_output)
    output_array = _get_output_array(out, a.shape, axis, func.keywords.get("keepdims", False), dtype, initial)

    where_compute = MaskWhere(True, a.shape)
    where_output = MaskWhere(True, output_array.shape)

    for index, chunk in a.iter_chunks(keepdims=True):
        _, chunk_selection, where_to_output = _get_indices(index, axis, where_compute, where_output, output_array.ndim)
        result = np.array(func(chunk), dtype=output_array.dtype)
        _apply_operation(operation, output_array, chunk_selection, where_to_output, result)

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array


def _get_str_dtype(
    a: npt.NDArray[np.str_] | H5Array[np.str_],
    b: npt.NDArray[Any] | H5Array[Any],
    func: NP_FUNC,
) -> np.dtype[Any]:
    assert np.issubdtype(a.dtype, str)

    if func == np.char.multiply:
        assert np.issubdtype(b.dtype, np.number)
        return np.dtype(f"<U{a.dtype.itemsize // 4 * max(b)}")

    if func == np.char.add:
        assert np.issubdtype(b.dtype, str)
        return np.dtype(f"<U{(a.dtype.itemsize + b.dtype.itemsize) // 4}")

    elif func in (
        np.char.greater,
        np.char.greater_equal,
        np.char.less,
        np.char.less_equal,
        np.char.equal,
        np.char.not_equal,
    ):
        return np.dtype(bool)

    raise NotImplementedError


def str_apply_2(
    func: NP_FUNC,
    a: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_],
    b: Any,
) -> npt.NDArray[Any] | H5Array[Any] | bool:
    a_arr = a if isinstance(a, (np.ndarray, ch5mpy.H5Array)) else np.array(a, dtype=str)
    b_arr = b if isinstance(b, (np.ndarray, ch5mpy.H5Array)) else np.array(b)

    if b_arr.dtype == object:
        b_arr = b_arr.astype(str)

    output_array = _get_output_array_2(None, a_arr.shape, b_arr.shape, _get_str_dtype(a_arr, b_arr, func), None)

    if not np.issubdtype(b_arr.dtype, str):
        try:
            return {np.char.equal: False, np.char.not_equal: True}[func]

        except KeyError:
            raise TypeError(f"'{func}' not supported between arrays with dtypes {a_arr.dtype} and {b_arr.dtype}.")

    for index, chunk_x1, chunk_x2 in iter_chunks_2(a_arr, b_arr):
        output_array[map_slice(index)] = func(chunk_x1, chunk_x2)

    return output_array


num_to_str_ufunc: dict[NP_FUNC, NP_FUNC] = {
    np.add: np.char.add,
    np.multiply: np.char.multiply,
    np.greater: np.char.greater,
    np.greater_equal: np.char.greater_equal,
    np.less: np.char.less,
    np.less_equal: np.char.less_equal,
    np.equal: np.char.equal,
    np.not_equal: np.char.not_equal,
}


def apply_2(
    func: NP_FUNC,
    a: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    b: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    *,
    out: H5Array[Any] | npt.NDArray[Any] | None,
    default: Any,
    dtype: npt.DTypeLike | None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | int | bool,
) -> Any:
    h5, maybe_not_h5, get_a, get_b = ensure_h5array_first(a, b)

    # if h5array is a str H5Array (got here because the operators ==, +, *, ... were used), pass to str_apply_2()
    if np.issubdtype(h5.dtype, str):
        assert out is None and where is True
        return str_apply_2(num_to_str_ufunc[func], a, b)  # type: ignore[arg-type]

    # operation on regular H5Arrays
    if not isinstance(maybe_not_h5, (np.ndarray, ch5mpy.H5Array)):
        maybe_not_h5 = np.array(maybe_not_h5)

    output_array = _get_output_array_2(out, h5.shape, maybe_not_h5.shape, dtype, default)

    if where is not False:
        where_compute = MaskWhere(where, get_a(h5, maybe_not_h5).shape)
        where_output = MaskWhere(where, output_array.shape)

        for index, chunk_x1, chunk_x2 in iter_chunks_2(get_a(h5, maybe_not_h5), get_b(h5, maybe_not_h5)):
            where_to_compute, chunk_selection, where_to_output = _get_indices(
                index, (), where_compute, where_output, output_array.ndim
            )
            result = np.array(
                func(
                    chunk_x1,
                    chunk_x2,
                    where=True if where_to_compute is None else where_to_compute,
                ),
                dtype=output_array.dtype,
            )
            _apply_operation(
                ApplyOperation.set,
                output_array,
                chunk_selection,
                where_to_output,
                result,
            )

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array



from __future__ import annotations

import importlib
from functools import partial
from typing import Callable

from ch5mpy._typing import H5_FUNC, NP_FUNC

HANDLED_FUNCTIONS: dict[NP_FUNC | NP_FUNC, H5_FUNC] = {}


def implements(*np_functions: NP_FUNC | NP_FUNC) -> Callable[[H5_FUNC], H5_FUNC]:
    """Register an __array_function__ implementation for H5Array objects."""

    def decorator(func: H5_FUNC) -> H5_FUNC:
        for f in np_functions:
            HANDLED_FUNCTIONS[f] = func
        return func

    return decorator


def register(function: partial[H5_FUNC], implements: NP_FUNC) -> None:
    HANDLED_FUNCTIONS[implements] = function


# manually import function implementations otherwise they are never imported
importlib.__import__("ch5mpy.array.functions.routines")
importlib.__import__("ch5mpy.array.functions.creation_routines")
importlib.__import__("ch5mpy.array.functions.two_arrays")
importlib.__import__("ch5mpy.array.functions.element_wise")
importlib.__import__("ch5mpy.array.functions.attributes")
importlib.__import__("ch5mpy.array.functions.sort")



from abc import ABC, abstractmethod
from typing import Any, Self

from scipy.sparse._base import sparray

from ch5mpy import Group, H5Dict


class H5_sparse_array(ABC, sparray):  # type: ignore[misc]
    # region magic methods
    @abstractmethod
    def __init__(
        self,
        file: Group | H5Dict[Any],
    ):
        self.file: H5Dict[Any] = H5Dict(file)

    def __repr__(self) -> str:
        return str("<H5 " + super().__repr__()[1:])

    @classmethod
    def __h5_read__(cls, values: H5Dict[Any]) -> Self:
        return cls(values)

    @abstractmethod
    def __h5_write__(self, values: H5Dict[Any]) -> None: ...

    # endregion

    # region attributes
    @property
    def _shape(self) -> tuple[int, int]:
        return tuple(self.file.attributes["_shape"])

    @property
    def shape(self) -> tuple[int, int]:
        return self._shape

    @shape.setter
    def shape(self, value: tuple[int, int]) -> None:
        self.file.attributes.set(_shape=value)

    # endregion

    # region class methods
    @classmethod
    def read(cls, file: Group | H5Dict[Any]) -> Self:
        return cls(file)

    @classmethod
    @abstractmethod
    def write(cls, file: Group | H5Dict[Any], matrix: sparray) -> Self: ...

    # endregion



from __future__ import annotations

from pathlib import Path
from typing import Protocol, runtime_checkable

import ch5mpy


@runtime_checkable
class AnonymousArrayCreationFunc(Protocol):
    def __call__(self, name: str, loc: str | Path | ch5mpy.File | ch5mpy.Group) -> None: ...



from __future__ import annotations

from dataclasses import dataclass
from itertools import chain, dropwhile, repeat, takewhile
from typing import Any, Generator, Iterable, Iterator, overload

import numpy as np
import numpy.typing as npt

from ch5mpy._typing import SELECTOR
from ch5mpy.indexing.base import Indexer, as_indexer, boolean_array_as_indexer
from ch5mpy.indexing.list import ListIndex
from ch5mpy.indexing.optimization import get_valid_indices
from ch5mpy.indexing.single import SingleIndex
from ch5mpy.indexing.slice import FullSlice
from ch5mpy.indexing.special import PLACEHOLDER, EmptyList, NewAxis, NewAxisType, PlaceHolderType
from ch5mpy.indexing.utils import takewhile_inclusive


@dataclass
class IndexedValue:
    value: Indexer
    idx: int


def get_indexer(
    obj: Indexer, sorted: bool = False, enforce_1d: bool = False, for_h5: bool = False
) -> int | npt.NDArray[np.int_] | slice | None:
    if isinstance(obj, NewAxisType):
        return None

    if isinstance(obj, (SingleIndex, FullSlice, EmptyList)):
        return obj.as_numpy_index()

    if isinstance(obj, ListIndex):
        if obj.size != 1:
            return obj.as_array(sorted=sorted, flattened=enforce_1d)

        if enforce_1d:
            return obj.squeeze().as_array().reshape((1,))

        if for_h5:
            return int(obj.squeeze().as_array()[()])

        return obj.squeeze().as_array()

    raise TypeError(f"Can't get indexer for {obj} with type '{type(obj)}'.")


def _compute_shape_empty_dset(
    indices: tuple[Indexer, ...], arr_shape: tuple[int, ...], new_axes: bool
) -> tuple[int, ...]:
    if new_axes and any(
        dropwhile(lambda x: x, reversed(list(dropwhile(lambda x: x, [x is NewAxis for x in indices]))))
    ):
        raise NotImplementedError("Cannot have new axis besides first and last indices yet.")

    shape: tuple[int, ...] = ()

    for index in indices:
        if index is NewAxis:
            if not new_axes:
                continue

            shape += (1,)
            continue

        if index.ndim > 0:
            shape += (len(index),)

    return shape + arr_shape[len(indices) :]


def _get_sorting_indices(i: Any) -> npt.NDArray[np.int_] | slice:
    if isinstance(i, np.ndarray):
        return np.unique(i.flatten(), return_inverse=True)[1]  # type: ignore[no-any-return]
    return slice(None)


def _as_indexers(indices: tuple[SELECTOR, ...], shape: tuple[int, ...]) -> Generator[Indexer, None, None]:
    if len([i for i in indices if i is not None]) > len(shape):
        raise IndexError(
            f"too many indices for H5Array, array is {len(shape)}-dimensional, but {len(indices)} were indexed"
        )

    shape_it = iter(shape)

    for index in indices:
        if index is None:
            yield NewAxis
            continue

        elif isinstance(index, (np.ndarray, list)):
            index = np.array(index)

            if index.dtype == np.bool_:
                for indexer in boolean_array_as_indexer(
                    index,  # type: ignore[arg-type]
                    tuple(next(shape_it) for _ in range(index.ndim)),
                ):
                    yield indexer

                continue

        yield as_indexer(index, next(shape_it))  # type: ignore[arg-type]


class Selection:
    __slots__ = "_indices", "_array_shape"

    # region magic methods
    def __init__(self, indices: Iterable[Indexer] | None, shape: tuple[int, ...], optimize: bool = True):
        if indices is None:
            self._indices: tuple[Indexer, ...] = ()
            return

        indices = tuple(indices)

        if len([i for i in indices if i is not NewAxis]) > len(shape):
            raise IndexError(f"Got too many indices ({len(indices)}) for shape {shape}.")

        it_indices = iter(indices)
        *_indices, first_list = takewhile_inclusive(lambda i: not isinstance(i, ListIndex), it_indices)

        largest_dim = 0 if first_list is None else max(i.ndim for i in indices if isinstance(i, ListIndex))
        self._indices = get_valid_indices(
            tuple(chain(_indices, () if first_list is None else (first_list.expand_to_dim(largest_dim),), it_indices)),
            optimize=optimize,
        )

        self._array_shape = shape

    def __repr__(self) -> str:
        return f"Selection{self._indices}\n         {self.in_shape} --> {self.out_shape}"

    @overload
    def __getitem__(self, item: int) -> Indexer: ...

    @overload
    def __getitem__(self, item: slice | Iterable[int]) -> Selection: ...

    def __getitem__(self, item: int | slice | Iterable[int]) -> Indexer | Selection:
        if isinstance(item, int):
            if item >= len(self._indices):
                if item >= len(self.out_shape):
                    raise IndexError(f"Index {item} is out of range for {self.out_shape} selection.")

                return FullSlice.whole_axis(self.out_shape[item])

            return self._indices[item]

        elif isinstance(item, slice):
            return Selection(self._indices[item], shape=self._array_shape)

        return Selection([e for i, e in enumerate(self._indices) if i in item], shape=self._array_shape)

    def __matmul__(self, item: slice | Iterable[int]) -> Selection:
        """Same as __getitem__ but without optimization"""
        if isinstance(item, slice):
            return Selection(self._indices[item], shape=self._array_shape, optimize=False)

        return Selection([e for i, e in enumerate(self._indices) if i in item], shape=self._array_shape, optimize=False)

    def __iter__(self) -> Iterator[Indexer]:
        return iter(self._indices)

    def __len__(self) -> int:
        return len(self._indices)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Selection):
            return False

        return (self._array_shape, self._indices) == (other._array_shape, other._indices)

    # endregion

    # region attributes
    @property
    def is_empty(self) -> bool:
        return len(self._indices) == 0

    @property
    def contains_empty_list(self) -> bool:
        return any(isinstance(e, EmptyList) for e in self._indices)

    @property
    def is_newaxis(self) -> npt.NDArray[np.bool_]:
        return np.array([x is NewAxis for x in self._indices], dtype=bool)

    @property
    def is_single(self) -> npt.NDArray[np.bool_]:
        return np.array([isinstance(x, SingleIndex) for x in self._indices], dtype=bool)

    @property
    def is_list(self) -> npt.NDArray[np.bool_]:
        return np.array([isinstance(x, (ListIndex, EmptyList)) for x in self._indices], dtype=bool)

    @property
    def is_slice(self) -> npt.NDArray[np.bool_]:
        return np.array([isinstance(x, FullSlice) for x in self._indices], dtype=bool)

    @property
    def in_shape(self) -> tuple[int, ...]:
        return self._array_shape

    @property
    def out_shape(self) -> tuple[int, ...]:
        if np.prod(self._array_shape) == 0:
            return _compute_shape_empty_dset(self._indices, self._array_shape, new_axes=True)

        len_end_shape = len(self._indices) - sum(self.is_newaxis)
        return self._min_shape(new_axes=True) + self._array_shape[len_end_shape:]

    @property
    def out_shape_squeezed(self) -> tuple[int, ...]:
        if np.prod(self.in_shape) == 0:
            shape = _compute_shape_empty_dset(self._indices, self._array_shape, new_axes=False)

        else:
            len_end_shape = len(self._indices) - sum(self.is_newaxis)
            shape = self._min_shape(new_axes=False) + self._array_shape[len_end_shape:]

        return tuple(
            s for s, is_list in zip(shape, chain(self.is_list | self.is_single, repeat(False))) if s != 1 or not is_list
        )

    # endregion

    # region methods
    @classmethod
    def from_selector(cls, indices: SELECTOR | tuple[SELECTOR, ...], shape: tuple[int, ...]) -> Selection:
        if not isinstance(indices, tuple):
            indices = (indices,)

        return Selection(_as_indexers(indices, shape), shape=shape)

    def get_indexers(
        self,
        sorted: bool = False,
        enforce_1d: bool = False,
        for_h5: bool = False,
    ) -> tuple[int | npt.NDArray[np.int_] | slice | None, ...]:
        """Get indexers for reading data in a h5py Dataset from a Selection.
        Args:
            sorted:
            enforce_1d:
            for_h5: skip new axes, transform lists of 1 element into single integer index
        """
        if for_h5:
            return tuple(
                get_indexer(i, sorted=sorted, enforce_1d=enforce_1d, for_h5=True)
                for i in self._indices
                if not isinstance(i, NewAxisType)
            )

        return tuple(get_indexer(i, sorted=sorted, enforce_1d=enforce_1d, for_h5=False) for i in self._indices)

    def _min_shape(self, new_axes: bool) -> tuple[int, ...]:
        last_slice_position = np.argwhere(self.is_slice)[-1, 0] if any(self.is_slice) else np.inf
        first_list_position = np.argmax(self.is_list) if any(self.is_list) else -1

        slice_indices_shape = tuple(len(s) for s in self._indices if isinstance(s, FullSlice))
        list_indices_shape = np.broadcast_shapes(
            *(lst.shape for lst in self._indices if isinstance(lst, (ListIndex, EmptyList)) and lst.ndim > 0)
        )

        if first_list_position < last_slice_position:
            min_shape = list_indices_shape + slice_indices_shape

        else:
            min_shape = slice_indices_shape + list_indices_shape

        if not new_axes:
            return min_shape

        if any(dropwhile(lambda x: x, reversed(list(dropwhile(lambda x: x, self.is_newaxis))))):
            raise NotImplementedError("Cannot have new axis besides first and last indices yet.")

        new_axes_before = tuple(1 for _ in takewhile(lambda x: x, self.is_newaxis))
        new_axes_after = tuple(1 for _ in takewhile(lambda x: x, reversed(self.is_newaxis[len(new_axes_before) :])))

        return new_axes_before + min_shape + new_axes_after

    def cast_on(self, previous_selection: Selection) -> Selection:
        if np.prod(previous_selection.out_shape) and self.in_shape != previous_selection.out_shape:
            raise ValueError(f"Cannot cast {self._array_shape} selection on {previous_selection.out_shape} selection.")

        if self.is_empty:
            return previous_selection

        if previous_selection.is_empty:
            return self

        previous_indices_queue = Queue(previous_selection._indices)
        self_indices_queue = Queue(self._indices)

        casted_selection: list[Indexer | PlaceHolderType] = [PLACEHOLDER] * len(previous_selection)

        while len(previous_indices_queue):
            if not len(self_indices_queue):
                one_previous_sel_element, idx = previous_indices_queue.next_one()
                casted_selection[idx] = one_previous_sel_element
                continue

            previous_sel_element = previous_indices_queue.next_all()

            if isinstance(previous_sel_element, IndexedValue):
                if isinstance(previous_sel_element.value, SingleIndex):
                    casted_selection[previous_sel_element.idx] = previous_sel_element.value
                    continue

                self_sel_element, _ = self_indices_queue.next_one()

                if isinstance(previous_sel_element.value, (FullSlice, EmptyList)):
                    casted_selection[previous_sel_element.idx] = previous_sel_element.value[self_sel_element]
                    continue

                assert isinstance(previous_sel_element.value, NewAxisType)
                if self_sel_element.as_numpy_index() != 0:
                    raise IndexError

                del casted_selection[previous_sel_element.idx]
                del previous_indices_queue[previous_sel_element.idx]

            else:
                self_sel_elements = tuple(e for (e, _) in self_indices_queue.next_n(previous_sel_element[0].value.ndim))

                for iv in previous_sel_element:
                    assert isinstance(iv.value, ListIndex)

                    casted_selection[iv.idx] = iv.value[self_sel_elements]

        casted_selection.extend(self_indices_queue)

        assert all(isinstance(e, Indexer) for e in casted_selection)

        sel = Selection(casted_selection, shape=previous_selection.in_shape)  # type: ignore[arg-type]

        if not sel.contains_empty_list and (sel.in_shape, sel.out_shape) != (
            previous_selection.in_shape,
            self.out_shape,
        ):
            raise RuntimeError("Selection casting failed: could not maintain shape consistency.")

        return sel

    def _get_loading_sel(self, extra_before: int = 0, extra_after: int = 0) -> tuple[int | slice, ...]:
        if extra_before == 0 and extra_after == 0 and not any(i is NewAxis for i in self._indices):
            return ()

        before = tuple(0 for _ in takewhile(lambda x: x is NewAxis, self._indices)) + (0,) * extra_before
        middle = tuple(
            slice(None)
            for i in takewhile(lambda x: x is not NewAxis, self._indices[len(before) :])
            if not isinstance(i, SingleIndex)
        )
        after = (
            tuple(0 for _ in takewhile(lambda x: x is NewAxis, self._indices[len(before) :][::-1])) + (0,) * extra_after
        )

        return before + middle + after

    def _get_loading_sel_iter(self, it: Iterator[int], list_indices: list[int]) -> Generator[int | slice, None, None]:
        yield from (0 for _ in takewhile(lambda i: i is NewAxis, self._indices))

        indices_it = enumerate(dropwhile(lambda i: i is NewAxis, self._indices))

        for _, s in dropwhile(lambda i_s: i_s[0] is NewAxis, zip(self._indices, self.out_shape)):
            if s == 1:
                yield 0

            else:
                idx, _ = next(indices_it)

                if idx in list_indices:
                    try:
                        yield next(it)
                    except StopIteration:
                        break

                else:
                    yield slice(None)

    def iter_indexers(
        self, can_reorder: bool = True
    ) -> Generator[
        tuple[
            tuple[int | npt.NDArray[np.int_] | slice | None, ...],
            slice | npt.NDArray[np.int_],
            tuple[int | npt.NDArray[np.int_] | slice, ...],
        ],
        None,
        tuple[npt.NDArray[np.int_] | slice, ...] | None,
    ]:
        list_indices = [idx for idx, i in enumerate(self._indices) if isinstance(i, ListIndex)]

        # short indexing ------------------------------------------------------
        if not len(list_indices):
            yield (
                self.get_indexers(for_h5=True),
                slice(None),
                self._get_loading_sel(),
            )
            return None

        if len(list_indices) == 1:
            list_ = self._indices[list_indices[0]]
            assert isinstance(list_, ListIndex)

            unique_list, inverse = np.unique(list_, return_inverse=True)

            already_sorted = np.array_equal(list_.as_array(sorted=True), list_.as_array())
            already_unique = len(unique_list) == len(list_)

            if can_reorder or already_sorted:
                if list_.ndim == 1:
                    indices = self.get_indexers(sorted=True, enforce_1d=True)

                    if already_unique:
                        yield (
                            indices,
                            slice(None),
                            self._get_loading_sel(),
                        )

                    else:
                        indices = indices[: list_indices[0]] + (unique_list,) + indices[list_indices[0] + 1 :]
                        yield (
                            indices,
                            inverse,
                            self._get_loading_sel(),
                        )
                        already_sorted = True

                elif list_.squeeze().ndim == 1:
                    extra_before = len(tuple(takewhile(lambda x: x == 1, list_.shape[:-1])))
                    extra_after = len(tuple(takewhile(lambda x: x, list_.shape[extra_before + 1 :][::-1])))

                    yield (
                        self.get_indexers(sorted=True, enforce_1d=True),
                        slice(None),
                        self._get_loading_sel(extra_before, extra_after),
                    )

                else:
                    raise NotImplementedError
                    # for idx, list_e in enumerate(list_.as_array(sorted=True)):
                    #     dataset_sel = tuple(
                    #         list_e if i == list_indices[0] else get_indexer(e, for_h5=True)
                    #         for i, e in enumerate(self._indices)
                    #     )
                    #     loading_sel = tuple(0 for s in self.out_shape if s == 1) + (idx,)

                    #     yield dataset_sel, loading_sel

                if not already_sorted:
                    sel_it = chain(self.get_indexers(for_h5=True), repeat(slice(None)))
                    return tuple(slice(None) if s == 1 else _get_sorting_indices(next(sel_it)) for s in self.out_shape)

                return None

        # long indexing -------------------------------------------------------
        if not self.out_shape_squeezed:
            yield (self.get_indexers(for_h5=True), slice(None), ())
            return None

        indices_dataset = [
            a.flatten() for a in np.broadcast_arrays(*(idx for i, idx in enumerate(self._indices) if i in list_indices))
        ]
        indices_loading = [i.flatten() for i in np.indices(self.out_shape_squeezed[: len(list_indices)])]

        for index_dataset, index_loading in zip(
            (iter(di) for di in zip(*indices_dataset)), (iter(li) for li in zip(*indices_loading))
        ):
            dataset_sel = tuple(
                next(index_dataset) if i in list_indices else get_indexer(e, for_h5=True)
                for i, e in enumerate(self._indices)
                if e is not NewAxis
            )
            loading_sel = tuple(self._get_loading_sel_iter(index_loading, list_indices))

            yield dataset_sel, slice(None), loading_sel

        return None

    # endregion


class Queue:
    # region magic methods
    def __init__(self, elements: Iterable[Indexer]):
        self._elements = list(elements)
        self._indices = list(range(len(self._elements)))

        self._broadcast_shape = np.broadcast_shapes(*[li.shape for li in self._elements if isinstance(li, ListIndex)])

    def __repr__(self) -> str:
        return f"Queue({self._elements})"

    def __len__(self) -> int:
        return len(self._elements)

    def __iter__(self) -> Iterator[Indexer]:
        return iter(self._elements)

    def __delitem__(self, index: int) -> None:
        if not len(self._indices):
            return

        if self._indices[0] <= index:
            raise ValueError

        self._indices = [i - 1 for i in self._indices]

    # endregion

    # region methods
    def next_one(self) -> tuple[Indexer, int]:
        if not len(self._elements):
            raise IndexError

        return self._elements.pop(0), self._indices.pop(0)

    def next_n(self, n: int) -> tuple[tuple[Indexer, int], ...]:
        if not len(self._elements):
            raise IndexError

        if (
            n == 1
            and len(self._elements) >= 2
            and isinstance(self._elements[0], FullSlice)
            and isinstance(self._elements[1], (ListIndex, EmptyList))
        ):
            return ((self._elements.pop(0), self._indices.pop(0)), (NewAxis, -1))

        return tuple((self._elements.pop(0), self._indices.pop(0)) for _ in range(min(n, len(self._elements))))

    def next_all(self) -> IndexedValue | tuple[IndexedValue, ...]:
        if not len(self._elements):
            raise IndexError

        if not isinstance(self._elements[0], ListIndex):
            return IndexedValue(self._elements.pop(0), self._indices.pop(0))

        lists = tuple(
            IndexedValue(e.broadcast_to(self._broadcast_shape), i)
            for e, i in zip(self._elements, self._indices)
            if isinstance(e, ListIndex)
        )

        self._indices = [i for e, i in zip(self._elements, self._indices) if not isinstance(e, ListIndex)]
        self._elements = [e for e in self._elements if not isinstance(e, ListIndex)]

        return tuple(li for li in lists)

    # endregion



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Collection
from typing import Any, Generic, Literal, TypeVar, cast, override

import h5py
import numpy as np
import numpy.typing as npt
from h5py.h5t import check_string_dtype
from numpy._typing import _ArrayLikeInt_co  # pyright: ignore[reportPrivateUsage]

import ch5mpy
from ch5mpy._typing import SELECTOR
from ch5mpy.attributes import AttributeManager
from ch5mpy.objects.pickle import PickleableH5Object
from ch5mpy.utils import NAN_PACKED

_T = TypeVar("_T", bound=np.generic)
_WT = TypeVar("_WT", bound=np.generic, covariant=True)
ENCODING = Literal["ascii", "utf-8"]
ERROR_METHOD = Literal[
    "backslashreplace",
    "ignore",
    "namereplace",
    "strict",
    "replace",
    "xmlcharrefreplace",
]


class DatasetWrapper(ABC, Generic[_WT]):
    """Base class to wrap Datasets."""

    # region magic methods
    def __init__(self, dset: Dataset[Any]):
        self._dset: Dataset[Any] = dset

    @abstractmethod
    def __getitem__(self, item: SELECTOR | tuple[SELECTOR, ...]) -> Any:
        pass

    def __setitem__(self, item: SELECTOR, value: Any) -> None:
        raise NotImplementedError

    def __getattr__(self, attr: str) -> Any:
        # If method/attribute is not defined here, pass the call to the wrapped dataset.
        return getattr(self._dset, attr)

    def __len__(self) -> int:
        return len(self._dset)

    # endregion

    # region numpy interface
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = True) -> npt.NDArray[Any]:
        if dtype is None:
            return np.array(self[()])

        return np.array(self[()]).astype(dtype)

    # endregion

    # region attributes
    @property
    @abstractmethod
    def dtype(self) -> np.dtype[Any]:
        pass

    @property
    def size(self) -> int:
        return self._dset.size

    @property
    def attrs(self) -> AttributeManager:
        return self._dset.attrs

    # endregion

    # region methods
    def read_direct(
        self,
        dest: npt.NDArray[Any],
        source_sel: tuple[int | slice | Collection[int], ...] | None = None,
        dest_sel: tuple[int | slice | Collection[int], ...] | None = None,
        expand_sel: _ArrayLikeInt_co | slice | None = None,
    ) -> None:
        source_sel = () if source_sel is None else source_sel
        dest_sel = () if dest_sel is None else dest_sel

        if expand_sel is None or isinstance(expand_sel, slice) and expand_sel == slice(None):
            dest[dest_sel] = self[source_sel]
        else:
            dest[dest_sel] = np.atleast_1d(self[source_sel])[expand_sel]

    # endregion


class AsStrWrapper(DatasetWrapper[np.str_]):
    """Wrapper to decode strings on reading the dataset"""

    # region magic methods
    @override
    def __getitem__(self, args: SELECTOR | tuple[SELECTOR, ...]) -> npt.NDArray[np.str_] | np.str_:
        subset = self._dset[args]

        if isinstance(subset, bytes):
            if subset == NAN_PACKED:
                return np.str_("nan")
            return np.str_(subset.decode())

        subset[np.where(subset == NAN_PACKED)] = b"nan"
        return np.array(subset, dtype=str)

    @override
    def __repr__(self) -> str:
        dset_name = "anonymous" if self._dset.name is None else self._dset.name[1:]
        return f'<HDF5 AsStrWrapper "{dset_name}": shape {self._dset.shape}">'

    # endregion

    # region attributes
    @property
    @override
    def dtype(self) -> np.dtype[np.str_]:
        str_dset = self._dset.asstr()[:]

        if isinstance(str_dset, np.ndarray):
            return str_dset.dtype

        return np.dtype(f"<U{len(str_dset)}")

    # endregion


class AsDtypeWrapper(DatasetWrapper[_WT]):
    """Wrapper to cast elements in a dataset as another numpy dtype."""

    # region magic methods
    def __init__(self, dset: Dataset[Any], dtype: npt.DTypeLike):
        super().__init__(dset)
        self._dtype: np.dtype[Any] = np.dtype(dtype)

    @override
    def __repr__(self) -> str:
        return f'<HDF5 AsDtypeWrapper "{self._dset.name[1:]}": shape {self._dset.shape}, dtype "{self._dtype}">'

    @override
    def __getitem__(self, args: SELECTOR | tuple[SELECTOR, ...]) -> npt.NDArray[Any] | Any:
        subset = self._dset[args]

        if np.isscalar(subset):
            return self._dtype.type(subset)  # type: ignore[arg-type]

        return np.array(subset, dtype=self._dtype)

    # endregion

    # region attributes
    @property
    @override
    def dtype(self) -> np.dtype[Any]:
        if np.issubdtype(self._dtype, str):
            # special case of str casting (todo: should find a better way of finding out the dtype)
            return self._dset[:].astype(str).dtype

        return self._dtype

    # endregion


class AsObjectWrapper(DatasetWrapper[_WT]):
    """Wrapper to map any object type to elements in a dataset."""

    # region magic methods
    def __init__(self, dset: Dataset[Any], otype: type[_WT]):
        super().__init__(dset)
        self._otype: type[_WT] = otype

    @override
    def __repr__(self) -> str:
        return (
            f'<HDF5 AsObjectWrapper "{self._dset.name[1:]}": shape {self._dset.shape}, otype "{self._otype.__name__}">'
        )

    @override
    def __getitem__(self, args: SELECTOR | tuple[SELECTOR, ...]) -> npt.NDArray[np.object_] | _WT:
        subset = self._dset[args]

        if np.isscalar(subset):
            return self._otype(subset)

        subset = cast(npt.NDArray[Any], subset)
        return np.array(list(map(self._otype, subset.flat)), dtype=np.object_).reshape(subset.shape)

    # endregion

    # region attributes
    @property
    @override
    def dtype(self) -> np.dtype[np.object_]:
        return np.dtype("O")

    @property
    def otype(self) -> type[_WT]:
        return self._otype

    # endregion


class Dataset(PickleableH5Object, h5py.Dataset, Generic[_T]):
    """A subclass of h5py.Dataset that implements pickling."""

    # region magic methods
    @override
    def __getitem__(
        self,
        arg: SELECTOR | tuple[SELECTOR, ...],
        new_dtype: npt.DTypeLike | None = None,
    ) -> _T | npt.NDArray[_T]:
        return super().__getitem__(arg, new_dtype)  # type: ignore[no-any-return]

    @override
    def __setitem__(self, arg: SELECTOR | tuple[SELECTOR, ...], val: Any) -> None:
        super().__setitem__(arg, val)

    # endregion

    # region attributes
    @property
    @override
    def file(self) -> ch5mpy.File:
        with h5py._objects.phil:  # type: ignore[attr-defined]
            return ch5mpy.File(self.id)

    @property
    def attributes(self) -> AttributeManager:
        return self.file.attrs

    @property
    @override
    def dtype(self) -> np.dtype[_T]:
        return self.id.dtype  # type: ignore[return-value]

    @property
    @override
    def attrs(self) -> AttributeManager:  # type: ignore[override]
        return AttributeManager(super().attrs)

    # endregion

    # region methods
    def asstr(self, encoding: ENCODING | None = None, errors: ERROR_METHOD = "strict") -> AsStrWrapper:  # type: ignore[override]
        """
        Get a wrapper to read string data as Python strings:

        The parameters have the same meaning as in `bytes.decode()`.
        If `encoding` is unspecified, it will use the encoding in the HDF5
        datatype (either ascii or utf-8).
        """
        string_info = check_string_dtype(self.dtype)
        if string_info is None:
            raise TypeError("dset.asstr() can only be used on datasets with an HDF5 string datatype")

        self_ = cast(Dataset[np.bytes_], self)
        return AsStrWrapper(self_)

    @override
    def astype(self, dtype: npt.DTypeLike) -> AsDtypeWrapper[np.generic]:  # type: ignore[override]
        return AsDtypeWrapper(self, dtype)

    def maptype(self, otype: type[Any]) -> AsObjectWrapper[Any]:
        # noinspection PyTypeChecker
        return AsObjectWrapper(self, otype)

    @override
    def read_direct(
        self,
        dest: npt.NDArray[Any],
        source_sel: tuple[int | slice | Collection[int], ...] | None = None,
        dest_sel: tuple[int | slice | Collection[int], ...] | None = None,
        expand_sel: _ArrayLikeInt_co | slice | None = None,
    ) -> None:
        # FIXME: find better way (here we have to load data to RAM before writing to dest)
        source_sel = () if source_sel is None else source_sel
        dest_sel = () if dest_sel is None else dest_sel

        if expand_sel is None or isinstance(expand_sel, slice) and expand_sel == slice(None):
            dest[dest_sel] = self[source_sel]
        else:
            dest[dest_sel] = np.atleast_1d(self[source_sel])[expand_sel]

    @override
    def write_direct(
        self,
        source: npt.NDArray[Any],
        source_sel: tuple[int | slice | Collection[int] | None, ...] | None = None,
        dest_sel: tuple[int | slice | Collection[int] | None, ...] | None = None,
    ) -> None:
        if not source.flags.carray:
            # ensure 'C' memory layout
            source = source.copy()

        if is_empty_selection(dest_sel):
            # empty selection : skip write operation
            return

        super().write_direct(handle_nan(source, self.dtype), source_sel=source_sel, dest_sel=dest_sel)

    # endregion


def is_empty_selection(sel: tuple[int | slice | Collection[int] | None, ...] | None) -> bool:
    if sel is None:
        return False

    if isinstance(sel, tuple) and len(sel) == 0:
        return False

    return not any(sel_len(s) for s in sel)


def sel_len(sel: int | slice | Collection[int] | None) -> int:
    if sel is None or isinstance(sel, (int, np.integer, slice)):
        return 1

    return len(sel)


def handle_nan(source: npt.NDArray[Any], dest_dtype: np.dtype[Any]) -> npt.NDArray[Any]:
    if np.issubdtype(dest_dtype, np.integer):
        if source.dtype.kind in "fc" and np.any(np.isnan(source)):
            raise ValueError("cannot convert float NaN to integer")

    if np.issubdtype(dest_dtype, np.object_):
        if source.ndim == 0:
            if source[()] is np.nan:
                return np.array(NAN_PACKED, dtype=object)

        else:
            is_nan = np.array([e is np.nan for e in source.ravel()], dtype=bool).reshape(source.shape)
            source[is_nan] = NAN_PACKED
            return source

    return source



from typing import Any, Callable, Generator, Iterable


def takewhile_inclusive(predicate: Callable[[Any], bool], it: Iterable[Any]) -> Generator[Any, None, None]:
    while True:
        e = next(it, None)  # type: ignore[call-overload]
        yield e

        if e is None or not predicate(e):
            break


def positive_slice_index(value: int, max: int) -> int:
    return value if value >= 0 else max + value



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np

from ch5mpy.array.functions.implement import implements

if TYPE_CHECKING:
    from ch5mpy.array.array import H5Array


@implements(np.ndim)
def ndim(a: H5Array[Any]) -> int:
    return a.ndim



from __future__ import annotations

import pickle
from collections.abc import ItemsView
from typing import TYPE_CHECKING, Any, override

import numpy as np
from h5py._hl.attrs import AttributeManager as H5Attrs
from h5py._hl.base import KeysViewHDF5

if TYPE_CHECKING:
    pass


_SPACER = ", \n                 "
_NONE_VALUE = "__h5_NONE__"


def _is_None(v: Any) -> bool:
    if isinstance(v, np.void):
        return False

    return isinstance(v, str) and v == _NONE_VALUE


class AttributeManager:
    """Manage attributes of a File, Group or Dataset."""

    # region magic methods
    def __init__(self, attrs: H5Attrs | AttributeManager) -> None:
        if isinstance(attrs, AttributeManager):
            self._attrs: H5Attrs = attrs._attrs

        else:
            self._attrs = attrs

    @override
    def __repr__(self) -> str:
        return f"AttributeManager{{{_SPACER.join([e[0] + ': ' + str(e[1]) for e in self.as_dict().items()])}}}"

    def __getitem__(self, name: str) -> Any:
        value = self._attrs[name]

        if isinstance(value, np.void):
            return pickle.loads(value.tobytes())

        if isinstance(value, str) and value == _NONE_VALUE:
            return None

        return value

    def __setitem__(self, name: str, value: Any) -> Any:
        if value is None:
            self._attrs[name] = _NONE_VALUE

        if isinstance(value, bytes):
            self._attrs[name] = np.void(value)

        else:
            try:
                self._attrs[name] = value

            except TypeError:
                # since <value> cannot be stored directly, we pickle it
                # here we use numpy's void type to allow storing bytes generated by pickle
                self._attrs[name] = np.void(pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL))

    def __delitem__(self, name: str) -> None:
        del self._attrs[name]

    def __len__(self) -> int:
        return len(self._attrs.keys())

    def __contains__(self, name: str) -> bool:
        return name in self._attrs.keys()

    # endregion

    # region attributes
    def keys(self) -> KeysViewHDF5[str]:
        return self._attrs.keys()

    def items(self) -> ItemsView[str, Any]:
        return self.as_dict().items()

    # endregion

    # region methods
    def get(self, name: str, default: Any = None) -> Any:
        return self._attrs.get(name, default)

    def set(self, **kwargs: Any) -> None:
        for k, v in kwargs.items():
            self[k] = v

    def as_dict(self) -> dict[str, Any]:
        return {k: None if _is_None(v) else v for k, v in self._attrs.items()}

    # endregion



from __future__ import annotations

from types import EllipsisType
from typing import Any, Callable, Iterable, SupportsIndex, Union

from numpy._typing import _ArrayLikeInt_co

ONE_AXIS_SELECTOR = Union[
    int,
    bool,
    SupportsIndex,
]

PARTIAL_SELECTOR = Union[
    None,
    EllipsisType,
    slice,
    range,
    Iterable[int],
    Iterable[bool],
    _ArrayLikeInt_co,
]

SELECTOR = Union[ONE_AXIS_SELECTOR, PARTIAL_SELECTOR]

NP_FUNC = Callable[..., Any]
H5_FUNC = Callable[..., Any]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, Literal, TypeVar

import numpy as np
import numpy.typing as npt

from ch5mpy.array.functions.implement import implements

if TYPE_CHECKING:
    from ch5mpy import H5Array


_T = TypeVar("_T", bound=np.generic)


@implements(np.copy)
def copy(a: H5Array[_T]) -> npt.NDArray[_T]:
    return np.array(a)


@implements(np.zeros_like)
def zeros_like(
    a: H5Array[Any],
    dtype: npt.DTypeLike | None = None,
    order: Literal["C", "F"] | None = "C",
    shape: Iterable[int] | None = None,
) -> Any:
    return np.zeros(shape=shape or a.shape, dtype=dtype or a.dtype, order=order)  # type: ignore[call-overload]


@implements(np.ones_like)
def ones_like(
    a: H5Array[Any],
    dtype: npt.DTypeLike = None,
    order: Literal["C", "F"] | None = "C",
    shape: Iterable[int] | None = None,
) -> Any:
    return np.ones(shape=shape or a.shape, dtype=dtype or a.dtype, order=order)  # type: ignore[call-overload]



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt

import ch5mpy


class ArrayCreationFunc:
    # region magic methods
    def __init__(self, name: str):
        self._name = name

    def __repr__(self) -> str:
        return f"<function ch5mpy.{self._name} at {hex(id(self))}>"

    def __call__(
        self,
        shape: int | tuple[int, ...],
        fill_value: Any,
        loc: str | Path | ch5mpy.File | ch5mpy.Group,
        name: str,
        *,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> ch5mpy.H5Array[Any]:
        shape = shape if isinstance(shape, tuple) else (shape,)

        if not isinstance(loc, ch5mpy.Group):
            loc = ch5mpy.File(loc, mode=ch5mpy.H5Mode.READ_WRITE_CREATE)

        dset = ch5mpy.store_dataset(
            None, loc, name, shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape, fill_value=fill_value
        )

        return ch5mpy.H5Array(dset)

    def defer(
        self,
        shape: int | tuple[int, ...],
        fill_value: Any,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> partial[ch5mpy.H5Array[Any]]:
        return partial(self.__call__, shape=shape, fill_value=fill_value, dtype=dtype, chunks=chunks, maxshape=maxshape)

    # endregion


class ArrayCreationFuncWithFill(ArrayCreationFunc):
    # region magic methods
    def __init__(self, name: str, fill_value: Any):
        super().__init__(name)
        self._fill_value = fill_value

    def __call__(  # type: ignore[override]
        self,
        shape: int | tuple[int, ...],
        loc: str | Path | ch5mpy.File | ch5mpy.Group,
        name: str,
        *,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> ch5mpy.H5Array[Any]:
        return super().__call__(
            loc=loc, name=name, shape=shape, fill_value=self._fill_value, dtype=dtype, chunks=chunks, maxshape=maxshape
        )

    # endregion

    # region methods
    def defer(  # type: ignore[override]
        self,
        shape: int | tuple[int, ...],
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> partial[ch5mpy.H5Array[Any]]:
        return partial(self.__call__, shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape)

    # endregion


full = ArrayCreationFunc("full")
empty = ArrayCreationFuncWithFill("empty", None)
zeros = ArrayCreationFuncWithFill("zeros", 0)
ones = ArrayCreationFuncWithFill("ones", 1)



from __future__ import annotations

from typing import Any, Generator, TypeVar, cast

import numpy as np
from numpy import typing as npt

from ch5mpy.indexing.selection import Selection
from ch5mpy.objects import Dataset, DatasetWrapper
from ch5mpy.utils import NAN_PACKED

_DT = TypeVar("_DT", bound=np.generic)


class IterWithFinalReordering:
    def __init__(self, gen: Generator[Any, Any, Any]):
        self.gen = gen
        self.final_reordering: tuple[npt.NDArray[np.int_] | slice, ...] | None = None

    def __iter__(self) -> Generator[Any, Any, Any]:
        self.final_reordering = yield from self.gen


def read_from_dataset(
    dataset: Dataset[_DT] | DatasetWrapper[_DT],
    selection: Selection,
    loading_array: npt.NDArray[_DT],
) -> None:
    if not dataset.size:
        if loading_array.size:
            raise ValueError("Reading from empty dataset.")
        return

    if not loading_array.size:
        return

    indexers = IterWithFinalReordering(selection.iter_indexers())
    for dataset_idx, dataset_expand_idx, loading_array_idx in indexers:
        # TODO : would be nice to be able to pass an array with random order in `dest_sel`
        dataset.read_direct(
            loading_array, source_sel=dataset_idx, dest_sel=loading_array_idx, expand_sel=dataset_expand_idx
        )

    if indexers.final_reordering is not None:
        loading_array[:] = loading_array[indexers.final_reordering]

    if loading_array.ndim == 0:
        if loading_array == NAN_PACKED:
            loading_array[()] = np.nan

    else:
        pass
        # is_nan = np.where(loading_array == NAN_PACKED)
        # if is_nan[0].size:
        #     loading_array[is_nan] = np.nan


def read_one_from_dataset(
    dataset: Dataset[_DT] | DatasetWrapper[_DT],
    selection: Selection,
    dtype: np.dtype[_DT],
) -> _DT:
    loading_array = np.empty((), dtype=dtype)
    read_from_dataset(dataset, selection, loading_array)
    return cast(_DT, loading_array[()])


def write_to_dataset(
    dataset: Dataset[_DT] | DatasetWrapper[_DT],
    values: npt.NDArray[_DT],
    selection: Selection,
) -> None:
    selection_shape = selection.out_shape

    if values.size == np.prod(selection_shape) and values.shape != selection_shape:
        values = values.reshape(selection_shape)

    for dataset_idx, _, array_idx in selection.iter_indexers(can_reorder=False):
        dataset.write_direct(values, source_sel=array_idx, dest_sel=dataset_idx)



from __future__ import annotations

import pickle
from typing import Any, cast, overload

import numpy as np
import numpy.typing as npt
from scipy.sparse import csr_array

from ch5mpy import Group
from ch5mpy.array import H5Array
from ch5mpy.dict import H5Dict
from ch5mpy.sparse._typing import INT_FLOAT
from ch5mpy.sparse.base import H5_sparse_array


class H5_csr_array(H5_sparse_array, csr_array):  # type: ignore[misc]
    """Compressed Sparse Row matrix created from hdf5 objects."""

    file: H5Dict[Any]
    data: H5Array[INT_FLOAT]
    indices: H5Array[np.integer[Any]]
    indptr: H5Array[np.integer[Any]]

    # region magic methods
    @overload
    def __new__(cls, arg1: Group | H5Dict[Any], *args: Any, **kwargs: Any) -> H5_csr_array: ...
    @overload
    def __new__(cls, arg1: tuple[int, int], *args: Any, **kwargs: Any) -> csr_array: ...
    def __new__(cls, arg1: Any, *args: Any, **kwargs: Any) -> csr_array:
        if not isinstance(arg1, (Group, H5Dict)):
            return csr_array(arg1, *args, **kwargs)

        return super(H5_csr_array, cls).__new__(cls)

    def __init__(
        self,
        file: Group | H5Dict[Any],
    ):
        super().__init__(file)

        self.data = self.file["data"]
        self.indices = self.file["indices"]
        self.indptr = self.file["indptr"]

    def __h5_write__(self, values: H5Dict[Any]) -> None:
        values["data"] = self.data
        values["indices"] = self.indices
        values["indptr"] = self.indptr

        values.attributes["_shape"] = self.shape

    # endregion

    # region class methods
    @classmethod
    def write(cls, file: Group | H5Dict[Any], matrix: csr_array) -> H5_csr_array:
        file = H5Dict(file)

        file["data"] = matrix.data
        file["indices"] = matrix.indices
        file["indptr"] = matrix.indptr

        file.attributes.set(
            _shape=matrix.shape,
            __h5_type__="object",
            __h5_class__=np.void(pickle.dumps(H5_csr_array, protocol=pickle.HIGHEST_PROTOCOL)),
        )

        return H5_csr_array(file)

    # endregion

    # region methods
    def _insert_many(
        self, i: npt.NDArray[np.integer[Any]], j: npt.NDArray[np.integer[Any]], x: npt.NDArray[INT_FLOAT]
    ) -> None:
        """Inserts new nonzero at each (i, j) with value x

        Here (i,j) index major and minor respectively.
        i, j and x must be non-empty, 1d arrays.
        Inserts each major group (e.g. all entries per row) at a time.
        Maintains has_sorted_indices property.
        """
        if self.has_sorted_indices:
            # convert i and j indices to a single ij array of complex values so that when we sort ij, i is used as
            # primary key and j as secondary
            sorted_indices = np.argsort(i + 1j * j)
            i = i[sorted_indices]
            j = j[sorted_indices]
            x = x[sorted_indices]

            ui, ui_idx = np.unique(i, return_index=True)

            split_indices = np.split(self.indices, cast(H5Array[np.int64], self.indptr[1:-1]))
            split_j = np.split(j, ui_idx[1:])

            indptr_diff = np.zeros(max(ui[-1] + 1, len(self.indptr) - 1), dtype=np.int64)
            indptr_diff[: len(self.indptr) - 1] = np.diff(self.indptr)
            n_items_before_line = np.r_[0, np.cumsum(indptr_diff)]

            def searchsorted(ix: int, jx: npt.NDArray[np.int64]) -> npt.NDArray[np.int64]:
                indices = np.searchsorted(split_indices[ix], jx) if ix < len(split_indices) else np.arange(len(jx))
                return cast(npt.NDArray[np.int64], indices + n_items_before_line[ix])

            insert_indices = np.concatenate([searchsorted(ix, jx) for ix, jx in zip(ui, split_j)])

            np.insert(self.data, insert_indices, x)
            np.insert(self.indices, insert_indices, j)

            indptr_diff[ui] += [len(s) for s in split_j]
            self.indptr.overwrite(np.cumsum(np.r_[0, indptr_diff]))

            self.shape = (max(self.shape[0], len(self.indptr) - 1)), max(self.shape[1], np.max(self.indices))

        else:
            raise NotImplementedError

    # endregion



from typing import Any

from ch5mpy.objects.dataset import Dataset, DatasetWrapper


class FlagDict:
    _flags: dict[str, bool]

    # region magic methods
    def __init__(self, dset: Dataset[Any] | DatasetWrapper[Any]) -> None:
        object.__setattr__(
            self,
            "_flags",
            {
                "C_CONTIGUOUS": dset.chunks is not None,
                "F_CONTIGUOUS": False,
                "OWNDATA": True,
                "WRITEABLE": True,
                "ALIGNED": False,  # TODO:
                "WRITEBACKIFCOPY": False,
            },
        )

    def __getitem__(self, key: str, /) -> bool:
        match key:
            case "FNC":
                return self._flags["F_CONTIGUOUS"] and not self._flags["C_CONTIGUOUS"]

            case "FORC":
                return self._flags["F_CONTIGUOUS"] or self._flags["C_CONTIGUOUS"]

            case "BEHAVED":
                return self._flags["ALIGNED"] and self._flags["WRITEABLE"]

            case "CARRAY":
                return self._flags["ALIGNED"] and self._flags["WRITEABLE"] and self._flags["C_CONTIGUOUS"]

            case "FARRAY":
                return (
                    self._flags["ALIGNED"]
                    and self._flags["WRITEABLE"]
                    and self._flags["F_CONTIGUOUS"]
                    and not self._flags["C_CONTIGUOUS"]
                )

            case _:
                return self._flags[key]

    def __setitem__(self, key: str, value: bool) -> None:
        assert key in self._flags.keys()
        self._flags[key] = value

    def __getattr__(self, name: str) -> bool:
        return self[name.upper()]

    def __setattr__(self, name: str, value: bool, /) -> None:
        self[name.upper()] = value

    # endregion



from __future__ import annotations

from typing import Any, override

import numpy as np
import numpy.typing as npt
from numpy._globals import _CopyMode

from ch5mpy.indexing.base import Indexer, as_indexer


class ListIndex(Indexer):
    # region magic methods
    def __init__(self, elements: npt.NDArray[np.int_], max: int):
        if not np.issubdtype(elements.dtype, np.integer):
            raise ValueError("Indexing elements should be integers.")
        if elements.ndim == 0:
            raise ValueError("Cannot build empty ListIndex, use EmptyList instead.")

        self._elements: npt.NDArray[np.int_] = np.array(elements)
        self._max: int = max

        self._elements[self._elements < 0] += max

        if len(self._elements) and (self._elements.min() < 0 or self._elements.max() >= max):
            raise IndexError(f"Selection {self._elements} is out of bounds for axis with size {max}.")

    @override
    def __repr__(self) -> str:
        flat_elements_repr = str(self._elements).replace("\n", "")
        return f"ListIndex({flat_elements_repr}, ndim={self.ndim} | {self._max})"

    def __getitem__(self, item: Indexer | tuple[Indexer, ...]) -> Indexer:
        if not isinstance(item, tuple):
            item = (item,)

        return as_indexer(self._elements[tuple(i.as_numpy_index() for i in item)], max=self._max)

    @override
    def __len__(self) -> int:
        return len(self._elements)

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ListIndex):
            return False

        return np.array_equal(self._elements, other._elements)

    @override
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | _CopyMode = True) -> npt.NDArray[Any]:
        if dtype is None:
            return self._elements

        return self._elements.astype(dtype, copy=copy)

    # endregion

    # region attributes
    @property
    @override
    def ndim(self) -> int:
        return self._elements.ndim

    @property
    def max(self) -> int:
        return self._max

    @property
    def shape(self) -> tuple[int, ...]:
        return self._elements.shape

    @property
    def size(self) -> int:
        return self._elements.size

    @property
    @override
    def is_whole_axis(self) -> bool:
        return np.array_equal(self._elements, np.arange(self._max))

    # endregion

    # region methods
    def as_array(self, sorted: bool = False, flattened: bool = False) -> npt.NDArray[np.int_]:
        elements = self._elements

        if flattened:
            elements = elements.flatten()

        if sorted:
            elements = np.sort(elements)

        return elements

    @override
    def as_numpy_index(self) -> npt.NDArray[np.int_]:
        return self.__array__()

    def squeeze(self) -> ListIndex:
        if self.size <= 1:
            return self

        return ListIndex(np.squeeze(self._elements), self._max)

    def expand_to_dim(self, n: int) -> ListIndex:
        if n < self.ndim:
            raise RuntimeError

        expanded_shape = (1,) * (n - self.ndim) + self.shape
        return ListIndex(self._elements.reshape(expanded_shape), self._max)

    def broadcast_to(self, shape: tuple[int, ...]) -> ListIndex:
        return ListIndex(np.broadcast_to(self._elements, shape), self._max)

    def reshape(self, shape: tuple[int, ...]) -> ListIndex:
        return ListIndex(self._elements.reshape(shape), self._max)

    def flatten(self) -> ListIndex:
        return ListIndex(self._elements.flatten(), self._max)

    # endregion



from __future__ import annotations

from typing import Any, Collection, Literal, cast

import h5py
import numpy.typing as npt
from h5py._hl.base import ItemsViewHDF5, ValuesViewHDF5

from ch5mpy.attributes import AttributeManager
from ch5mpy.names import H5Mode
from ch5mpy.objects.dataset import Dataset
from ch5mpy.objects.pickle import PickleableH5Object


def _h5py_wrap_type(obj: Any) -> Any:
    """Produce our objects instead of h5py default objects"""
    if isinstance(obj, h5py.Dataset):
        return Dataset(obj.id)
    elif isinstance(obj, h5py.File):
        return File(obj.id)
    elif isinstance(obj, h5py.Group):
        return Group(obj.id)
    elif isinstance(obj, h5py.Datatype):
        return obj  # Not supported for pickling yet. Haven't really thought about it
    else:
        return obj  # Just return, since we want to wrap h5py.Group.get too


class Group(PickleableH5Object, h5py.Group):
    """
    A subclass of h5py.Group that implements pickling, and to create new groups and datasets
    of the right type (i.e. the ones defined in this module).
    """

    # region magic methods
    def __getitem__(self, name: str | bytes) -> Any:  # type: ignore[override]
        return self._wrap(h5py.Group.__getitem__(self, name))  # type: ignore[index]

    # endregion

    # region attributes
    @property
    def attrs(self) -> AttributeManager:  # type: ignore[override]
        return AttributeManager(super().attrs)

    @property
    def file(self) -> File:
        with h5py._objects.phil:  # type: ignore[attr-defined]
            return File(self.id)

    @property
    def parent(self) -> Group:
        return Group(super().parent.id)

    # endregion

    # region methods
    def _wrap(self, obj: Any) -> Any:
        """Wrap an object accessed in this group with our custom classes."""
        obj = _h5py_wrap_type(obj)

        # If it is a group or dataset copy the current file info in
        if isinstance(obj, Group) or isinstance(obj, Dataset):
            obj.file_info = self.file_info

        return obj

    def get(self, name: str, default: Any = None, getclass: bool = False, getlink: bool = False) -> Any:
        """Retrieve an item or other information.

        "name" given only:
            Return the item, or "default" if it doesn't exist

        "getclass" is True:
            Return the class of object (Group, Dataset, etc.), or "default"
            if nothing with that name exists

        "getlink" is True:
            Return HardLink, SoftLink or ExternalLink instances.  Return
            "default" if nothing with that name exists.

        "getlink" and "getclass" are True:
            Return HardLink, SoftLink and ExternalLink classes.  Return
            "default" if nothing with that name exists.

        Example:

        >>> cls = group.get('foo', getclass=True)
        >>> if cls == SoftLink:
        """
        if not getclass and not getlink:
            return self._wrap(h5py.Group.get(self, name, default))  # type: ignore[arg-type]

        return h5py.Group.get(self, name, default, getclass, getlink)  # type: ignore[call-overload]

    def require_group(self, name: str) -> Group:
        """
        Return a group, creating it if it doesn't exist.

        TypeError is raised if something with that name already exists that
        isn't a group.
        """
        return cast(Group, super().require_group(name))

    def values(self) -> ValuesViewHDF5[Group | Dataset[Any]]:  # type: ignore[override]
        return super().values()  # type: ignore[return-value]

    def items(self) -> ItemsViewHDF5[str, Group | Dataset[Any]]:  # type: ignore[override]
        return super().items()  # type: ignore[return-value]

    def create_group(self, name: str, track_order: bool | None = None, overwrite: bool = False) -> Group:
        """
        Create and return a new subgroup.

        Args:
            name: may be absolute or relative. Fails if the target name already exists.
            track_order: Track dataset/group/attribute creation order under this group if True. If None use global
                default h5.get_config().track_order.
            overwrite: overwrite group if it already exists ?
        """
        if overwrite and name in self.keys():
            del self[name]

        group = super().create_group(name, track_order=track_order)
        return cast(Group, self._wrap(group))

    def create_dataset(
        self,
        name: str | None,
        shape: int | tuple[()] | tuple[int | None, ...] | None = None,
        dtype: npt.DTypeLike | None = None,
        data: Collection[Any] | None = None,
        **kwds: Any,
    ) -> Dataset[Any]:
        """
        Create and return a new Dataset.

        Args:
            name: Name of the dataset (absolute or relative). Provide None to make an anonymous dataset.
            shape: Dataset shape. Use "()" for scalar datasets. Required if "data" isn't provided.
            dtype: Numpy dtype or string. If omitted, dtype('f') will be used. Required if "data" isn't provided;
                otherwise, overrides data array's dtype.
            data: Provide data to initialize the dataset. If used, you can omit shape and dtype arguments.
            kwds: other arguments to pass to the dataset creation function.
        """
        group = super().create_dataset(name, shape=shape, dtype=dtype, data=data, **kwds)
        return cast(Dataset[Any], self._wrap(group))

    # endregion


class File(Group, h5py.File):
    """A subclass of h5py.File that implements pickling."""

    # region magic methods
    def __init__(self, *args: Any, **kwargs: Any):
        # Store args and kwargs for pickling
        self.init_args = args
        self.init_kwargs = kwargs

    def __new__(cls, *args: Any, **kwargs: Any) -> File:
        """Create a new File object with the h5 open function."""
        with h5py._objects.phil:  # type: ignore[attr-defined]
            self = super().__new__(cls)
            h5py.File.__init__(self, *args, **kwargs)

            return self

    def __getstate__(self) -> None:
        pass

    def __getnewargs_ex__(self) -> tuple[tuple[Any, ...], dict[str, Any]]:
        kwargs = self.init_kwargs.copy()

        if len(self.init_args) > 1 and self.init_args[1] == "w":
            return (self.init_args[0], "r+", *self.init_args[2:]), kwargs

        return self.init_args, kwargs

    def __enter__(self) -> File:
        return super().__enter__()  # type: ignore[return-value]

    # endregion

    # region attributes
    @property
    def mode(self) -> Literal[H5Mode.READ, H5Mode.READ_WRITE]:  # type: ignore[override]
        return H5Mode(super().mode)  # type: ignore[return-value]

    @property
    def file_info(self) -> File:
        return self

    # endregion



from itertools import chain
from typing import cast

import numpy as np

from ch5mpy.indexing.base import Indexer
from ch5mpy.indexing.list import ListIndex
from ch5mpy.indexing.single import SingleIndex
from ch5mpy.indexing.slice import FullSlice
from ch5mpy.indexing.special import EmptyList, NewAxis
from ch5mpy.indexing.utils import takewhile_inclusive


class Vector:
    # region magic methods
    def __init__(self, index: ListIndex):
        self._index = index
        self._max = index.max

    def __repr__(self) -> str:
        return f"Vector({self._index})"

    # endregion

    # region methods
    def try_as_slice(self) -> FullSlice | None:
        array = self._index.as_array(flattened=True)
        array[array < 0] += self._index.max

        if not len(array):
            return None

        if len(array) == 1:
            return FullSlice.one(array[0], max=self._index.max)

        steps = np.ediff1d(array)

        # check all steps are equal and > 0
        if steps[0] <= 0 or np.any(steps != steps[0]):
            return None

        return FullSlice(array[0], array[-1] + steps[0], steps[0], max=self._index.max)

    def drop_axes(self, n: int) -> ListIndex:
        n = min(self._index.ndim - 1, n)

        return cast(ListIndex, self._index[(SingleIndex(0, max=self._max),) * n])

    # endregion


class VectorEnsembl:
    # region magic methods
    def __init__(self, vectors: tuple[Vector, ...]):
        self._vectors = vectors

    # endregion

    # region methods
    def convert(self) -> tuple[FullSlice | ListIndex, ...]:
        converted_vectors: tuple[FullSlice | ListIndex, ...] = ()
        can_convert = True
        drop_n_axes = 0

        for vector in self._vectors:
            if can_convert:
                converted = vector.try_as_slice()
                if converted is None:
                    can_convert = False
                    converted_vectors += (vector.drop_axes(drop_n_axes),)

                else:
                    drop_n_axes += 1
                    converted_vectors += (converted,)

            else:
                converted_vectors += (vector.drop_axes(drop_n_axes),)

        return converted_vectors

    # endregion


def get_valid_indices(indices: tuple[Indexer, ...], optimize: bool) -> tuple[Indexer, ...]:
    if not len(indices):
        return ()

    if any(isinstance(i, EmptyList) for i in indices):
        # make sure shapes are broadcastable
        np.broadcast_shapes(*(lst.shape for lst in indices if isinstance(lst, (ListIndex, EmptyList)) and lst.ndim > 0))

    if not optimize:
        return indices

    if _gets_whole_dataset(indices):
        return ()

    if any(isinstance(i, ListIndex) for i in indices) and all(isinstance(i, (FullSlice, ListIndex)) for i in indices):
        indices = _optimize_indices(indices)

    if _gets_whole_dataset(indices):
        return ()

    return _drop_whole_axes(indices)


def _gets_whole_dataset(indices: tuple[Indexer, ...]) -> bool:
    return all(isinstance(i, (FullSlice, ListIndex)) and i.is_whole_axis for i in indices)


def _drop_whole_axes(indices: tuple[Indexer, ...]) -> tuple[Indexer, ...]:
    drop = 0
    rev_indices = list(indices[::-1])

    for idx, i in enumerate(rev_indices):
        if isinstance(i, FullSlice) and i.is_whole_axis:
            drop += 1

        elif isinstance(i, ListIndex) and i.squeeze().is_whole_axis:
            if len(rev_indices) == (idx + 1):
                drop += 1
                continue

            next_index = rev_indices[idx + 1]

            if not isinstance(next_index, ListIndex) or next_index.ndim == 1:
                drop += 1

            elif next_index.ndim == 2 and next_index.shape[1] == 1:
                rev_indices[idx + 1] = next_index.flatten()
                drop += 1

            else:
                break

        else:
            break

    return tuple(rev_indices[drop:][::-1])


def _convert_lists_to_slices(indices: tuple[ListIndex, ...]) -> tuple[Indexer, ...]:
    """
    From indexing arrays, return (if possible) the tuple of slices that would lead to the same indexing since indexing
    from slices leads to much fewer read operations.
    Also, return the number of NewAxes to add before and after the tuple of slices so as to maintain the resulting
    array's dimensions.

    This is effectively the inverse of the numpy.ix_ function, with indexing arrays converted to slices.
    """
    if len(indices) == 1 and len(indices[0]) == 1:
        return indices

    ndim = max(i.ndim for i in indices)

    if len(indices) > ndim:
        raise ValueError

    shapes = np.array([i.expand_to_dim(ndim).shape for i in indices])

    start = ndim - len(indices)
    end = start + len(indices)
    extra_before, square, extra_after = np.split(shapes, [start, end], axis=1)

    square[np.diag_indices(len(indices))] = 1

    if np.any(extra_before != 1) or np.any(square != 1) or np.any(extra_after != 1):
        raise ValueError

    slices = VectorEnsembl(tuple(Vector(i) for i in indices)).convert()

    return (NewAxis,) * extra_before.shape[1] + slices + (NewAxis,) * extra_after.shape[1]


def _optimize_indices(indices: tuple[Indexer, ...]) -> tuple[Indexer, ...]:
    list_indices, lists = zip(*((idx, i) for idx, i in enumerate(indices) if isinstance(i, ListIndex)))

    try:
        lists = _convert_lists_to_slices(lists)
    except ValueError:
        pass

    it_optimized_lists = iter(lists)

    return tuple(
        chain.from_iterable(
            (i,) if idx not in list_indices else takewhile_inclusive(lambda e: e is NewAxis, it_optimized_lists)
            for idx, i in enumerate(indices)
        )
    )



from __future__ import annotations

from typing import Any, Literal, override

import numpy as np
import numpy.typing as npt

from ch5mpy.indexing.base import Indexer


class SingleIndex(Indexer):
    __slots__: tuple[str, ...] = "_index", "_max"

    # region magic methods
    def __init__(self, index: int, max: int):
        if max <= index:
            raise IndexError(f"Selection {index} is out of bounds for axis with size {max}.")

        self._index: int = index
        self._max: int = max

    @override
    def __repr__(self) -> str:
        return f"SingleIndex({self._index} | {self._max})"

    @override
    def __len__(self) -> int:
        return 1

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SingleIndex):
            return False

        return self._index == other._index

    @override
    def __array__(self, dtype: npt.DTypeLike | None = None, copy: bool | None = True) -> npt.NDArray[Any]:
        return np.array(self._index, dtype=None)

    # endregion

    # region attributes
    @property
    @override
    def ndim(self) -> Literal[0]:
        return 0

    @property
    @override
    def is_whole_axis(self) -> bool:
        return self._max == 1

    @property
    def max(self) -> int:
        return self._max

    # endregion

    # region methods
    @override
    def as_numpy_index(self) -> int:
        return self._index

    def as_slice(self) -> slice:
        return slice(self._index, self._index + 1)

    # endregion



from typing import Union

from ch5mpy.indexing.base import Indexer, as_indexer, boolean_array_as_indexer
from ch5mpy.indexing.list import ListIndex
from ch5mpy.indexing.selection import Selection, get_indexer
from ch5mpy.indexing.single import SingleIndex
from ch5mpy.indexing.slice import FullSlice, map_slice
from ch5mpy.indexing.special import EmptyList, NewAxis, NewAxisType

LengthedIndexer = Union[ListIndex, FullSlice, EmptyList]

__all__ = [
    "Indexer",
    "Selection",
    "get_indexer",
    "boolean_array_as_indexer",
    "map_slice",
    "FullSlice",
    "ListIndex",
    "SingleIndex",
    "NewAxis",
    "NewAxisType",
    "as_indexer",
    "EmptyList",
]
